﻿using System;
using System.Collections;
using System.Collections.Generic;
using System.Linq;
using UnityEngine;
using SimpleJSON; // JSONNode

namespace JustAnotherUser {
    public class TriggerIncrementer : MVRScript, TriggerIncrementerActions {
        private List<DAZMorph> _morphs, _usedMorphs;
        private List<SceneObject> _colliders; // the object that will colide with the collider region
        private MotionAnimationControl _head;
        private DAZCharacterSelector _characterSelector;

        private TriggerIncrementerUI _ui;

        private float _duration;

        private float _finalTime = 0f; // play the animation until finalTime
        private Atom _collisionBox; // area where to detect collision

        private static readonly string COLLISION_BOX_NAME = "TI_collision_{plugin_id}_{object_id}";
        private static readonly float COLLISION_BOX_SIZE = 0.4f;

        private static readonly string VERSION = "1.1";

        public override void Init() {
            // plugin VaM GUI description
            pluginLabelJSON.val = "TriggerIncrementer v" + VERSION;

            // get the objects
            if (containingAtom.type != "Person" || (this._characterSelector = containingAtom.GetComponentInChildren<DAZCharacterSelector>()) == null) {
                throw new InvalidOperationException("Missing DAZCharacterSelector");
            }

            if ((this._head = GetHead()) == null) throw new InvalidOperationException("Head not found in added object");

            this._ui = new TriggerIncrementerUI(this, this);

            this._morphs = new List<DAZMorph>();
            this._usedMorphs = new List<DAZMorph>();

            this._ui.InitUI();
        }


        // Runs once when plugin loads (after Init)
        protected void Start() {
            this._morphs.Clear();
            ScanBank(this._characterSelector.morphBank1, this._morphs); // @author https://github.com/ProjectCanyon/morph-merger/blob/master/MorphMerger.cs
            ScanBank(this._characterSelector.morphBank2, this._morphs);
            ScanBank(this._characterSelector.morphBank3, this._morphs);

            StartCoroutine(GetCollisionBox(COLLISION_BOX_NAME.Replace("{plugin_id}", this.storeId).Replace("{object_id}", containingAtom.name)));

            //SuperController.LogMessage(pluginLabelJSON.val + " Loaded");

            this._ui.LoadJson(GetPluginJsonFromSave());
        }

        private void SetupCollisionBox() {
            this._collisionBox.hidden = true;

            JSONStorable trigger = this._collisionBox.GetStorableByID("Trigger");
            JSONClass triggerJSON = trigger.GetJSON();

            if (triggerJSON["trigger"]["startActions"].AsArray.Count == 0) {
                triggerJSON["trigger"]["startActions"][0].Add("receiverAtom", "");
                triggerJSON["trigger"]["startActions"][0].Add("receiver", "");
                triggerJSON["trigger"]["startActions"][0].Add("receiverTargetName", "");
                triggerJSON["trigger"]["startActions"][0].Add("boolValue", "");
            }
            triggerJSON["trigger"]["startActions"][0]["receiverAtom"].Value = this.containingAtom.name;
            triggerJSON["trigger"]["startActions"][0]["receiver"].Value = this.storeId; // plugin ID
            triggerJSON["trigger"]["startActions"][0]["receiverTargetName"].Value = "run";
            triggerJSON["trigger"]["startActions"][0]["boolValue"].Value = "true";
            trigger.LateRestoreFromJSON(triggerJSON); // TODO what will happend with the older object? will it keep the JSON?
        }

        public void OnCollision() {
            List<Atom> collidingAtoms = this.GetDesiredAtomsColliding();
            if (collidingAtoms.Count == 0) return;

            foreach (Atom collide in collidingAtoms) collide.Remove();

            float incrementTime = this._duration * collidingAtoms.Count;
            if (this._finalTime > Time.time) this._finalTime += incrementTime;
            else this._finalTime = Time.time + incrementTime;
        }
        
        public static List<Atom> GetCollidingAtoms(CollisionTrigger trigger) {
            CollisionTriggerEventHandler handler = trigger.GetComponentInChildren<CollisionTriggerEventHandler>();

            List<Atom> r = new List<Atom>();
            if (handler == null) return r;

            foreach (KeyValuePair<Collider,bool> collider in handler.collidingWithDictionary) {
                Atom collided = SuperController.singleton.GetAtoms().FirstOrDefault(a =>
                        a.GetComponentsInChildren<Collider>().FirstOrDefault(c => c == collider.Key) != null);
                if (collided != null) r.Add(collided);
            }
            return r;
        }

        public List<Atom> GetDesiredAtomsColliding() {
            if (this._collisionBox == null) return new List<Atom>();

            List<Atom> colliding = TriggerIncrementer.GetCollidingAtoms(this._collisionBox.GetComponentInChildren<CollisionTrigger>()); // TODO use collision box wrapper
            colliding.RemoveAll(a => !this.IsDesired(a));
            return colliding;
        }

        public bool IsDesired(Atom a) {
            if (this._colliders.Count < 1) return false;
            return this._colliders.Any(obj => obj.Equals(a.name));
        }

        public void SetTargetObjects(List<SceneObject> objects) {
            this._colliders = objects;

            //SuperController.LogMessage("New collider");
        }

        public void AnimationDurationChanged(float newDuration) {
            this._duration = newDuration;
        }

        public void Update() {
            if (this._collisionBox == null) return; // wait to load
            this._collisionBox.mainController.transform.position = this._head.transform.TransformPoint(Vector3.down * 0.05f);

            if (this._finalTime < Time.time) return;
            float secondsSinceLastUpdate = Time.deltaTime;
            foreach (DAZMorph morph in this._usedMorphs) {
                float addPerAnimation = this._ui.GetMorphIncrement(morph.displayName),
                    addPerSecond = addPerAnimation / this._duration;
                morph.SetValue(morph.appliedValue + addPerSecond*secondsSinceLastUpdate);
                morph.SyncJSON();
            }
            // TODO check GetFloatJSONParamMaxValue (?)
        }


        public List<string> GetAllMorphNames() {
            return this._morphs.Select(m => m.displayName).Distinct().ToList();
        }

        private IEnumerator GetCollisionBox(string name) {
            // does it already exists?
            this._collisionBox = SuperController.singleton.GetAtoms().FirstOrDefault(a => a.name == name);
            if (this._collisionBox == null) {
                // no collision box; generate a new one
                SuperController.LogMessage("Generating collision box...");

                yield return SuperController.singleton.AddAtomByType("CollisionTrigger", name);
                this._collisionBox = SuperController.singleton.GetAtoms().FirstOrDefault(a => a.name == name);

                // modify the new collision box
                this._collisionBox.GetStorableByID("scale").GetFloatJSONParam("scale").val = COLLISION_BOX_SIZE;
            }

            this.SetupCollisionBox();
        }

        public void AddMorph(string morphName) {
            DAZMorph morph = FindMorphByName(this._morphs, morphName);
            if (morph == null) {
                SuperController.LogError("Unable to find morph '" + morphName + "'");
                return;
            }

            this._usedMorphs.Add(morph);
        }

        // @author https://raw.githubusercontent.com/ChrisTopherTa54321/VamScripts/master/FloatMultiParamRandomizer.cs
        public JSONNode GetPluginJsonFromSave() {
            foreach (JSONNode atoms in SuperController.singleton.loadJson["atoms"].AsArray) {
                if (!atoms["id"].Value.Equals(containingAtom.name)) continue;

                foreach (JSONNode storable in atoms["storables"].AsArray) {
                    if (storable["id"].Value == this.storeId) {
                        return storable;
                    }
                }
            }

            return null;
        }

        private MotionAnimationControl GetHead() {
            foreach (MotionAnimationControl mac in containingAtom.motionAnimationControls) { // TODO get head inside linkableRigidbodies?
                if (!mac.name.Equals("headControl")) continue;
                return mac;
            }

            return null; // not found
        }

        private void ScanBank(DAZMorphBank bank, List<DAZMorph> morphs) { // TODO only morph (not morph & pose)
            if (bank == null) return;

            foreach (DAZMorph morph in bank.morphs) {
                if (!morph.visible) continue;

                morphs.Add(morph);
                //SuperController.LogMessage(morph.morphName);
            }
        }

        private DAZMorph FindMorphByName(List<DAZMorph> morphs, string name) {
            foreach (DAZMorph morph in morphs) {
                if (!morph.displayName.Equals(name)) continue;

                return morph;
            }

            return null; // not found
        }
    }
}