﻿using System;
using System.Collections;
using System.Collections.Generic;
using System.Linq;
using UnityEngine;
using SimpleJSON; // JSONNode
using UnityEngine.UI; // InputField

namespace JustAnotherUser {
    public class TriggerIncrementerUI {
        private MVRScript _script;
        private TriggerIncrementerActions _ti;

        private JSONStorableBool _activeStorable;
        private JSONStorableFloat _durationStorable;

        private IDictionary<string, JSONStorableFloat> _morphIncrement; // all the morphs that it should modify and its increment
        private List<JSONStorableParam> _colliderSelectors;

        public TriggerIncrementerUI(TriggerIncrementer script, TriggerIncrementerActions ti) {
            this._script = script;
            this._ti = ti;

            this._morphIncrement = new Dictionary<string, JSONStorableFloat>();
            this._colliderSelectors = new List<JSONStorableParam>();
        }

        // @ref https://hub.virtamate.com/threads/coding-canvas-and-layers-in-vam-text-inputs.2839/
        private static UIDynamicTextField CreateTextInput(MVRScript script, JSONStorableString jss, bool rightSide = false) {
            var textfield = script.CreateTextField(jss, rightSide);
            textfield.height = 20f;
            textfield.backgroundColor = Color.white;
            var input = textfield.gameObject.AddComponent<InputField>();
            var rect = input.GetComponent<RectTransform>().sizeDelta = new Vector2(1f, 0.4f);
            input.textComponent = textfield.UItext;
            jss.inputField = input;
            return textfield;
        }

        public void InitUI() {
            UIDynamicButton addLiteralBtn = this._script.CreateButton("Add literal object");
            addLiteralBtn.button.onClick.AddListener(() => this.AddLiteralCollider());

            UIDynamicButton addRegexBtn = this._script.CreateButton("Add regex object");
            addRegexBtn.button.onClick.AddListener(() => this.AddRegexCollider());
            
            this._durationStorable = new JSONStorableFloat("duration", 10.0f, (float val) => this._ti.AnimationDurationChanged(val), 0.1f, 120.0f);
            this._script.RegisterFloat(this._durationStorable);
            this._script.CreateSlider(this._durationStorable, true); // create on the right

            this._activeStorable = new JSONStorableBool("run", false, (val) => {
                if (!val) return;

                this._ti.OnCollision();
                this._activeStorable.val = false;
            });
            this._script.RegisterBool(this._activeStorable); // we don't care about this information, but we need to register it to set it on the trigger
            this._script.CreateToggle(this._activeStorable, true); // create it on the right

            JSONStorableStringChooser morphChooseList = new JSONStorableStringChooser("morph", null, "", "Add morph", (string morphName) => {
                SuperController.LogMessage("Added " + morphName);

                this.AddMorphSlider(morphName);
                this._ti.AddMorph(morphName);
            });
            // we don't care about this information
            var linkPopup = this._script.CreateFilterablePopup(morphChooseList, true);
            linkPopup.popupPanelHeight = 600f;
            linkPopup.popup.onOpenPopupHandlers += () => { morphChooseList.choices = this.GetUnusedMorphNames(); };
        }

        public void AddLiteralCollider(string name = "") {
            JSONStorableStringChooser colliderStorable = new JSONStorableStringChooser("collider_literal_" + this._colliderSelectors.Count().ToString(),
                        null, "", "Collider", (string colliderName) => this.ColliderStorableChanged());
            this._script.RegisterStringChooser(colliderStorable);
            var literalLinkPopup = this._script.CreateFilterablePopup(colliderStorable);
            literalLinkPopup.popupPanelHeight = 600f;
            literalLinkPopup.popup.onOpenPopupHandlers += () => { colliderStorable.choices = SuperController.singleton.GetAtoms().Select(a => a.name).Distinct().ToList(); }; // TODO subtract already added
            
            this._colliderSelectors.Add(colliderStorable);

            colliderStorable.val = name;
        }

        public void AddRegexCollider(string name = "") {
            JSONStorableString colliderStorable = new JSONStorableString("collider_regex_" + this._colliderSelectors.Count().ToString(),
                        "", (string newValue) => this.ColliderStorableChanged());
            this._script.RegisterString(colliderStorable);
            TriggerIncrementerUI.CreateTextInput(this._script, colliderStorable);

            this._colliderSelectors.Add(colliderStorable);

            colliderStorable.val = name;
        }
        
        public void LoadJson(JSONNode node) {
            if (node == null) return;

            foreach (string entry in node.AsObject.Keys) {
                switch (entry) {
                    case "collider":
                    case "duration":
                        break; // automatically loaded

                    case "run":
                        this._activeStorable.val = false; // for some reason, sometimes it saves as true (thus not working)
                        break;

                    case "id":
                    case "pluginLabel":
                        break; // ignore

                    default:
                        if (entry.StartsWith("collider_literal_")) {
                            this.AddLiteralCollider(node[entry]);
                        }
                        else if (entry.StartsWith("collider_regex_")) {
                            this.AddRegexCollider(node[entry]);
                        }
                        else {
                            // tracked morphs
                            try {
                                this.AddMorphSlider(entry, node[entry].AsFloat);
                                this._ti.AddMorph(entry);
                            } catch (Exception ex) { }
                        }
                        break;
                }
            }

            this._ti.AnimationDurationChanged(this._durationStorable.val);
            this.ColliderStorableChanged();
        }

        private void ColliderStorableChanged() {
            List<SceneObject> colliders = new List<SceneObject>();
            foreach (JSONStorableParam storable in this._colliderSelectors) {
                if (storable.GetType().Equals(typeof(JSONStorableStringChooser))) {
                    // Literal
                    string value = ((JSONStorableStringChooser)storable).val;
                    if (!value.Equals("")) colliders.Add(new ConcreteSceneObject(value));
                }
                else if (storable.GetType().Equals(typeof(JSONStorableString))) {
                    // RegEx
                    string value = ((JSONStorableString)storable).val;
                    if (!value.Equals("")) {
                        try {
                            colliders.Add(new MultipleSceneObject(value));
                        } catch (Exception ex) {
                            SuperController.LogError("Error while parsing RegEx '" + value + "': " + ex.Message);
                        }
                    }
                }
                else SuperController.LogError("Unknown class " + storable.GetType());
            }

            this._ti.SetTargetObjects(colliders);
        }

        public void AddMorphSlider(string morphName, float value = 0.0f) {
            JSONStorableFloat jsonFloat = new JSONStorableFloat(morphName, 0.0f, -1.0f, 1.0f);
            jsonFloat.val = value;
            this._script.RegisterFloat(jsonFloat);
            this._script.CreateSlider(jsonFloat, true);

            this._morphIncrement.Add(morphName, jsonFloat);
        }

        public float GetMorphIncrement(string morphName) {
            try {
                return this._morphIncrement[morphName].val;
            } catch (KeyNotFoundException) {
                return 0.0f; // no morph => no increment
            }
        }
        
        public List<string> GetSelectedMorphNames() {
            return this._morphIncrement.Keys.ToList();
        }

        private List<string> GetUnusedMorphNames() {
            List<string> r = this._ti.GetAllMorphNames(),
                        alreadySelected = this.GetSelectedMorphNames();
            r.RemoveAll(s => alreadySelected.Contains(s));
            return r;
        }
    }
}